from abc import ABCMeta, abstractmethod

import json
import networkx as nx
from matchescu.typing import EntityReference, EntityReferenceIdentifier


class GraphPersistence(metaclass=ABCMeta):
    @abstractmethod
    def load(self) -> nx.Graph:
        pass

    @abstractmethod
    def save(self, g: nx.Graph) -> None:
        pass


class GmlGraphPersistence(GraphPersistence):
    def __init__(self, file_path):
        super().__init__()
        self._path = file_path

    @staticmethod
    def _json_decode(dct: dict):
        obj_type = dct.pop("type", "ref_id")
        if obj_type == "ref":
            ref_id = dct.pop("ref_id")
            ref = EntityReference(ref_id, dct)
            return ref
        if obj_type == "ref_id":
            return EntityReferenceIdentifier(dct.get("label"), dct.get("source"))
        return dct

    def load(self) -> nx.Graph:
        return nx.read_gml(
            self._path,
            destringizer=lambda x: json.loads(x, object_hook=self._json_decode),
        )

    @staticmethod
    def _json_encode(o):
        if isinstance(o, EntityReferenceIdentifier):
            return {"type": "ref_id", "label": o.label, "source": o.source}
        if isinstance(o, EntityReference):
            return {
                "type": "ref",
                "ref_id": {"label": o.id.label, "source": o.id.source},
                **o.as_dict(),
            }
        return o

    def save(self, g: nx.Graph) -> None:
        return nx.write_gml(
            g, self._path, stringizer=lambda x: json.dumps(x, default=self._json_encode)
        )



from typing import Protocol, TypeVar

from matchescu.typing import EntityReference


TRef = TypeVar("TRef", bound=EntityReference)


class Matcher(Protocol[TRef]):
    @property
    def non_match_threshold(self) -> float:
        """Similarity score (ranged 0..1) below which two references are considered to truly mismatch."""
        pass

    @property
    def match_threshold(self) -> float:
        """Similarity score (ranged 0..1) above which two references are considered to truly match one another."""
        pass

    def __call__(self, left: TRef, right: TRef) -> float:
        """Return a similarity score between ``left`` and ``right``.

        :param left: an entity reference of any kind
        :param right: an entity reference of any kind

        :return: a ``float`` value ranged between 0 and 1 which represents the
            probability that ``left`` matches ``right``.
        """
        raise NotImplementedError()



from dataclasses import dataclass
from typing import Hashable, Iterable, Sized, Protocol, Callable, Any, runtime_checkable

from matchescu.typing._data import Record


@dataclass(frozen=True, eq=True)
class EntityReferenceIdentifier:
    """Identifies an entity reference.

    Attributes:
        :label Hashable: unique label identifying an entity reference within a collection
        :source str: a string describing where the entity reference originated
    """

    label: Hashable
    source: str

    def __repr__(self) -> str:
        return f"ref_id{{{self.source},{repr(self.label)}}}"

    def __str__(self) -> str:
        return f"{self.source}({str(self.label)})"


@runtime_checkable
class EntityReference(Record, Protocol):
    """An entity reference instance allows accessing data by name or index.

    Attributes:
        id EntityReferenceIdentifier: identifies the entity reference
    """

    id: EntityReferenceIdentifier

    def as_dict(self) -> dict[str, Any]:
        """Return a dictionary containing the reference's attribute names and values."""
        pass


@runtime_checkable
class EntityProfile(Iterable[EntityReference], Sized, Protocol):
    """Entity profiles are entity references aggregating other entity references.

    Concrete entity profiles have different representations according to the
    entity resolution model being used. However, all entity profiles can be
    treated as 'compound' entity references obtained by aggregating existing
    'atomic' entity references.

    Just like any other entity reference, an entity profile has a unique ID.
    However, because they were generated by the entity resolution process, they
    have no source.
    """


EntityReferenceIdFactory = Callable[[Iterable[Record]], EntityReferenceIdentifier]



from matchescu.similarity._matcher import Matcher
from matchescu.similarity._persistence import GraphPersistence, GmlGraphPersistence
from matchescu.similarity._reference_graph import ReferenceGraph


__all__ = ["GraphPersistence", "GmlGraphPersistence", "Matcher", "ReferenceGraph"]



from typing import Generator, Generic

import networkx as nx

from matchescu.similarity._matcher import Matcher, TRef
from matchescu.similarity._persistence import GraphPersistence
from matchescu.typing._references import EntityReferenceIdentifier


class ReferenceGraph(Generic[TRef]):
    """Graph representation of the similarity between entity references.

    The nodes of the graph are entity references that were compared against each
    other when the ``add`` method was called. Edges in this graph are weighted
    using the matcher passed to the constructor.

    Reference graphs may be directed or undirected. By default, they are
    undirected, suggesting that the matcher is symmetric,
    i.e. ``matcher(x, y) == matcher(y, x)``.
    """

    def __init__(
        self,
        matcher: Matcher[TRef],
        directed: bool = False,
    ) -> None:
        self.__directed = directed
        self.__g = nx.DiGraph() if directed else nx.Graph()
        self.__matcher = matcher

    def __repr__(self):
        return "SimilarityGraph(nodes={}, edges={}, matcher={})".format(
            len(self.__g.nodes),
            len(self.__g.edges),
            repr(self.__matcher),
        )

    @property
    def directed(self):
        return self.__directed

    @property
    def nodes(self):
        """Returns the nodes of the graph."""
        return self.__g.nodes

    @property
    def edges(self):
        """Returns the edges of the graph along with their similarity weights and types."""
        return self.__g.edges(data=True)

    def add(self, left: TRef, right: TRef) -> "ReferenceGraph":
        """Add an edge between two entity references.

        The edge is added based on the configured similarity thresholds based
        on the similarity computed by the configured matcher.

        :param left: left entity reference
        :param right: right entity reference

        :return: ``self``, with the added edge.
        """
        sim_score = self.__matcher(left, right)
        self.__g.add_edge(left.id, right.id, weight=sim_score, refs=(left, right))
        return self

    def matches(
        self, match_min: float = 0.75
    ) -> Generator[
        tuple[EntityReferenceIdentifier, EntityReferenceIdentifier], None, None
    ]:
        yield from (
            (u, v)
            for u, v, weight in self.__g.edges.data("weight", default=0.0)
            if weight >= match_min
        )

    def potential_matches(
        self, non_match_max: float = 0.25, match_min: float = 0.75
    ) -> Generator[
        tuple[EntityReferenceIdentifier, EntityReferenceIdentifier], None, None
    ]:
        yield from (
            (u, v)
            for u, v, weight in self.__g.edges.data("weight", default=0.0)
            if non_match_max <= weight < match_min
        )

    def non_matches(
        self, non_match_max: float = 0.25
    ) -> Generator[
        tuple[EntityReferenceIdentifier, EntityReferenceIdentifier], None, None
    ]:
        yield from (
            (u, v)
            for u, v, weight in self.__g.edges.data("weight", default=0.0)
            if weight < non_match_max
        )

    def has_edge(
        self, u: EntityReferenceIdentifier, v: EntityReferenceIdentifier
    ) -> bool:
        return (u, v) in self.__g.edges

    def weight(
        self, left: EntityReferenceIdentifier, right: EntityReferenceIdentifier
    ) -> float:
        data = self.__g.get_edge_data(left, right, default={})
        return float(data.get("weight", 0.0))

    def load(self, persistence: GraphPersistence) -> "ReferenceGraph":
        self.__g = persistence.load()
        self.__directed = self.__g.is_directed()
        return self

    def save(self, persistence: GraphPersistence) -> "ReferenceGraph":
        persistence.save(self.__g)
        return self

    def to_undirected(self) -> "ReferenceGraph":
        """Convert the graph to an undirected graph.

        If the graph is already undirected, this method returns a copy of the
        graph.
        """
        other = ReferenceGraph(self.__matcher, directed=False)
        if self.__directed:
            other.__g.add_edges_from(self.__g.edges(data=True))
        else:
            other.__g = self.__g.copy()
        return other

    def to_directed(self) -> "ReferenceGraph":
        """Convert the graph to a bidirectional directed graph.

        If the graph is already directed, this method returns a copy of the
        graph.
        """
        other = ReferenceGraph(self.__matcher, directed=True)
        if self.__directed:
            other.__g = self.__g.copy()
        else:
            for _, __, (u, v) in self.__g.edges.data("refs", default=0.0):
                other.add(u, v)
                other.add(v, u)
        return other

    def merge(self, other: "ReferenceGraph") -> "ReferenceGraph":
        if self.__matcher != other.__matcher:
            raise ValueError("Cannot merge graphs with different matchers.")
        if self.__directed != other.__directed:
            raise ValueError("Cannot merge graphs with different directions.")
        g = nx.compose(self.__g, other.__g)
        result = ReferenceGraph(self.__matcher, directed=self.__directed)
        result.__g = g
        return result
